#ifndef _MP3_INDEX_H_
#define _MP3_INDEX_H_

#include "../../WDL/fileread.h"

class mp3_index
{
public:
  mp3_index(const char *fn) { m_refcnt=0; m_numframes=m_framestart=0; m_framefile=0; m_fn.Set(fn); m_start_eatsamples=0; m_end_eatsamples=0; }
  ~mp3_index() { delete m_framefile; m_framefile=0; }

  int GetFrameCount() { return m_framefile||m_numframes==m_frameposmemcache.GetSize()?m_numframes:0; }
  int GetFramePos(int f) 
  { 
    WDL_MutexLock lock(&m_mutex);
    if (m_framefile && m_numframes > 0)
    {
      m_framefile->SetPosition(m_framestart + f * 4);
      unsigned char buf[4];
      if (m_framefile->Read(buf,4) == 4)
        return buf[0] | (buf[1]<<8) | (buf[2]<<16) | (buf[3]<<24);        
    }
    else if (m_numframes==m_frameposmemcache.GetSize() && f>=0&&f<m_numframes) return m_frameposmemcache.Get()[f];

    return 0;
  }

  static mp3_index *indexFromFilename(const char *fn);
  static void release_index(mp3_index *idx)
  {
    WDL_MutexLock lock(&indexMutex);
    if (--idx->m_refcnt<=0)
    {
      int x;
      for (x = 0; x < g_indexes.GetSize() && g_indexes.Get(x) != idx; x ++);
      if (x < g_indexes.GetSize())
        g_indexes.Delete(x);

      delete idx;
    }
  }

private:
  int m_refcnt;
  WDL_Mutex m_mutex;
  WDL_String m_fn;
  static WDL_Mutex indexMutex;
  static WDL_PtrList<mp3_index> g_indexes;
  static int _sortfunc(const void *a, const void *b);

  int ReadFrameListFromCache(); // 0 if found
  void BuildFrameList();
  int m_numframes;
  int m_framestart;
  WDL_FileRead *m_framefile;

  WDL_TypedQueue<unsigned int> m_frameposmemcache; // only used if !m_framefile, and then only if the mp3 is small enough

public:
  int m_start_eatsamples;
  int m_end_eatsamples;
};


#endif