#ifdef _WIN32
#include <windows.h>
#else
#include "../../WDL/swell/swell.h"
#endif
#include <stdio.h>
#include <math.h>
#include <sys/stat.h>
#include "../../WDL/lameencdec.h"

#include "resource.h"

#include "../reaper_plugin.h"

#include "../../WDL/LineParse.h"
#include "../../WDL/wdlstring.h"
#include "../../WDL/filewrite.h"


extern REAPER_PeakBuild_Interface *(*PeakBuild_Create)(PCM_source *src, const char *fn, int srate, int nch);
extern void (*update_disk_counters)(int read, int write);
extern const char *(*get_ini_file)();
extern HWND g_main_hwnd;

typedef enum 
{
	LQP_NOPRESET=-1,

	// QUALITY PRESETS
	LQP_NORMAL_QUALITY		= 0,
	LQP_LOW_QUALITY			= 1,
	LQP_HIGH_QUALITY		= 2,
	LQP_VOICE_QUALITY		= 3,
	LQP_R3MIX				= 4,
	LQP_VERYHIGH_QUALITY	= 5,
	LQP_STANDARD			= 6,
	LQP_FAST_STANDARD		= 7,
	LQP_EXTREME				= 8,
	LQP_FAST_EXTREME		= 9,
	LQP_INSANE				= 10,
	LQP_ABR					= 11,
	LQP_CBR					= 12,
	LQP_MEDIUM				= 13,
	LQP_FAST_MEDIUM			= 14, 

	// NEW PRESET VALUES
	LQP_PHONE	=1000,
	LQP_SW		=2000,
	LQP_AM		=3000,
	LQP_FM		=4000,
	LQP_VOICE	=5000,
	LQP_RADIO	=6000,
	LQP_TAPE	=7000,
	LQP_HIFI	=8000,
	LQP_CD		=9000,
	LQP_STUDIO	=10000

} LAME_QUALTIY_PRESET;

extern HINSTANCE g_hInst;
#define WIN32_FILE_IO


static void FixSampleRate(int &sr)
{
  if (sr <= 8000) sr=8000;
  else if (sr <= 11025) sr=11025;
  else if (sr <= 12000) sr=12000;
  else if (sr <= 16000) sr=16000;
  else if (sr <= 22050) sr=22050;
  else if (sr <= 24000) sr=24000;
  else if (sr <= 32000) sr=32000;
  else if (sr <= 44100) sr=44100;
  else sr=48000;
}


class PCM_sink_mp3lame : public PCM_sink
{
  public:
    static HWND showConfig(void *cfgdata, int cfgdata_l, HWND parent);

    PCM_sink_mp3lame(const char *fn, void *cfgdata, int cfgdata_l, int nch, int srate, bool buildpeaks)
    {
        m_peakbuild=0;
        m_bitrate = 128;
        m_stereomode = 1; //joint stereo
        m_quality = 0; //normal
        m_vbrmethod = -1; //no vbr
        m_vbrq = 2;
        m_abr = 128;
        m_vbrmax = 320;

        if (cfgdata_l >= 32 && *((int *)cfgdata) == REAPER_FOURCC('m','p','3','l'))
        {
          m_bitrate=REAPER_MAKELEINT(((int *)(((unsigned char *)cfgdata)+4))[0]);
          m_stereomode=REAPER_MAKELEINT(((int *)(((unsigned char *)cfgdata)+4))[1]);
          m_quality=REAPER_MAKELEINT(((int *)(((unsigned char *)cfgdata)+4))[2]);
          m_vbrmethod=REAPER_MAKELEINT(((int *)(((unsigned char *)cfgdata)+4))[3]);
          m_vbrq=REAPER_MAKELEINT(((int *)(((unsigned char *)cfgdata)+4))[4]);
          m_vbrmax=REAPER_MAKELEINT(((int *)(((unsigned char *)cfgdata)+4))[5]);
          m_abr=REAPER_MAKELEINT(((int *)(((unsigned char *)cfgdata)+4))[6]);
        }

        FixSampleRate(srate);

        m_nch=nch>1?2:1;
        m_srate=srate;
        m_lensamples=0;
        m_filesize=0;
        m_fn.Set(fn);
        m_enc=0;
        m_fh=new WDL_FileWrite(fn,false);
        if (m_fh && !m_fh->IsOpen())
        {
          delete m_fh;
          m_fh=0;
        }
        if (m_fh) m_enc=new LameEncoder(srate,nch,m_bitrate,m_stereomode,m_quality,m_vbrmethod,m_vbrq, m_vbrmax, m_abr);

        if (m_enc)
          m_enc->SetVBRFilename(m_fn.Get());

        if (buildpeaks && m_enc)
        {
          m_peakbuild=PeakBuild_Create(NULL,fn,m_srate,m_nch);
        }

    }

    bool IsOpen()
    {
      return m_enc && m_fh;
    }

    ~PCM_sink_mp3lame()
    {
      if (IsOpen())
      {
        m_enc->Encode(NULL,0,1);

        FlushOut();
        // flush output

      }

      delete m_fh;
      m_fh=0;

      delete m_enc; // be sure to delete m_enc AFTER m_fh, to ensure it can write the vbr tag etc
      m_enc=0;

      delete m_peakbuild;
      m_peakbuild=0;
    }

    const char *GetFileName() { return m_fn.Get(); }
    int GetNumChannels() { return m_nch; } // return number of channels
    double GetLength() { return m_lensamples / (double) m_srate; } // length in seconds, so far
    INT64 GetFileSize()
    {
      return m_filesize;
    }
    int GetLastSecondPeaks(int sz, ReaSample *buf)
    {
      if (m_peakbuild)
        return m_peakbuild->GetLastSecondPeaks(sz,buf);
      return 0;
    }
    void GetPeakInfo(PCM_source_peaktransfer_t *block)
    {
      if (m_peakbuild) m_peakbuild->GetPeakInfo(block);
      else block->peaks_out=0;
    }

    void GetOutputInfoString(char *buf, int buflen)
    {
      char tmp[512];
      sprintf(tmp,"MP3 (lame_enc) %dHz %dch, %dkbps",m_srate,
          m_enc?m_enc->GetNumChannels() : m_nch,
          m_bitrate);
      lstrcpyn(buf,tmp,buflen);
    }

    void WriteMIDI(MIDI_eventlist *events, int len, double samplerate) { }
    void WriteDoubles(ReaSample **samples, int len, int nch, int offset, int spacing)
    {
      if (m_peakbuild)
        m_peakbuild->ProcessSamples(samples,len,nch,offset,spacing);

      ReaSample *tmpptrs[2]={samples[0]+offset,samples[0]+offset};     
      if (nch > 1) tmpptrs[1]=samples[1]+offset;

      float *tmpbuf=m_inbuf.Resize(len*m_nch,false);
      float *p=tmpbuf;
      int splsout=0;
      while (len-->0)
      {          
        int ch;
        for (ch = 0; ch < m_nch; ch ++)
        {
          ReaSample v=tmpptrs[ch][0];
          if (v <= -1.0) *p++=-1.0f;
          else if (v >= 1.0) *p++=1.0f;
          else *p++=(float)v;
          tmpptrs[ch]+=spacing;
          splsout++;
        }
      }
      if (splsout)
      {
        m_lensamples+=splsout/m_nch;
        m_enc->Encode(tmpbuf,splsout/m_nch,1);

        FlushOut();
      }
    }

    void FlushOut()
    {
      int l=m_enc->outqueue.Available();
      if (l>0)
      {
        m_fh->Write(m_enc->outqueue.Get(),l);
        if (update_disk_counters) update_disk_counters(0,l);
        m_filesize+=l;
        m_enc->outqueue.Advance(l);
        m_enc->outqueue.Compact();
      }
    }
    int Extended(int call, void *parm1, void *parm2, void *parm3) 
    {
      if (call == PCM_SINK_EXT_VERIFYFMT)
      {
        if (parm1 && parm2)
        {
          if (*(int *)parm2 <= 1) *(int *)parm2=1;
          else *(int *)parm2 = 2;

          int sr=*(int *)parm1;

          FixSampleRate(sr);

          *(int *)parm1 = sr;

          return 1;
        }
      }
      return 0;
    }


 private:
    WDL_FileWrite *m_fh;
    int m_bitrate;
    int m_vbrq, m_abr, m_vbrmax, m_quality, m_stereomode, m_vbrmethod;

    WDL_TypedBuf<float> m_inbuf;
    int m_nch,m_srate;
    INT64 m_filesize;
    INT64 m_lensamples;
    WDL_String m_fn;
    LameEncoder *m_enc;
    REAPER_PeakBuild_Interface *m_peakbuild;
};

static unsigned int GetFmt(char **desc) 
{
  if (desc) *desc="MP3 (encoder by LAME project)";
  return REAPER_FOURCC('m','p','3','l');
}

static const char *GetExtension(const void *cfg, int cfg_l)
{
  if (cfg_l >= 4 && *((int *)cfg) == REAPER_FOURCC('m','p','3','l')) return "mp3";
  return NULL;
}


static int m_bitrates[] = { 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 0 };

static void setBitrates(HWND hwndDlg, int mi, int ma)
{
  for(int i=0;m_bitrates[i];i++)
  {
    if (m_bitrates[i] == mi) SendDlgItemMessage(hwndDlg,IDC_BITRATE,CB_SETCURSEL,i,0);
    if (m_bitrates[i] == ma) SendDlgItemMessage(hwndDlg,IDC_MAXBITRATE,CB_SETCURSEL,i,0);
  }
}


static void SetQualityStr(HWND hwndDlg, const char* txt, int idx)
{
  int n = SendDlgItemMessage(hwndDlg, IDC_QUALITY, CB_GETCOUNT, 0, 0);
  SendDlgItemMessage(hwndDlg, IDC_QUALITY, CB_ADDSTRING, n, (LPARAM)txt);
  SendDlgItemMessage(hwndDlg, IDC_QUALITY, CB_SETITEMDATA, n, idx);
}

static void SetQuality(HWND hwndDlg, int idx)
{
  int i, n = SendDlgItemMessage(hwndDlg, IDC_QUALITY, CB_GETCOUNT, 0, 0);
  for (i = 0; i < n; ++i)
  {
    int j = SendDlgItemMessage(hwndDlg, IDC_QUALITY, CB_GETITEMDATA, i, 0);
    if (j == idx) 
    {
      SendDlgItemMessage(hwndDlg, IDC_QUALITY, CB_SETCURSEL, i, 0);
      break;
    }
  }
}

static int GetQuality(HWND hwndDlg)
{
  int i = SendDlgItemMessage(hwndDlg, IDC_QUALITY, CB_GETCURSEL, 0, 0);
  if (i == CB_ERR) return 6;  // preset standard
  return SendDlgItemMessage(hwndDlg, IDC_QUALITY, CB_GETITEMDATA, i, 0);
}

static void updateOptions(HWND hwndDlg)
{
  int vbr_meth=SendDlgItemMessage(hwndDlg,IDC_VBRMETHOD,CB_GETCURSEL,0,0);
  if (vbr_meth == CB_ERR) vbr_meth = 0;
  vbr_meth--;
  
  bool dovbrq = (vbr_meth != -1 && vbr_meth != 4);
  ShowWindow(GetDlgItem(hwndDlg,IDC_VBRQ), dovbrq ? SW_SHOW : SW_HIDE);
  ShowWindow(GetDlgItem(hwndDlg,IDC_VBRQ_TEXT), dovbrq ? SW_SHOW : SW_HIDE);
  //EnableWindow(GetDlgItem(hwndDlg,IDC_VBRQ),vbr_meth != -1 && vbr_meth != 4);
  //EnableWindow(GetDlgItem(hwndDlg,IDC_VBRQ_TEXT),vbr_meth != -1 && vbr_meth != 4);


  SetDlgItemText(hwndDlg,IDC_TEXT1,vbr_meth==-1?"kbps:":vbr_meth==4?"Min kbps:":"Min kbps:");
  SetDlgItemText(hwndDlg,IDC_TEXT2,vbr_meth==-1?"":vbr_meth==4?"Max kbps:":"Max kbps:");
  SetDlgItemText(hwndDlg,IDC_TEXT3,vbr_meth!=4?"":"Avg kbps:");

  ShowWindow(GetDlgItem(hwndDlg,IDC_MAXBITRATE),vbr_meth != -1?SW_NORMAL:SW_HIDE);
  ShowWindow(GetDlgItem(hwndDlg,IDC_AVGBITRATE),vbr_meth == 4?SW_NORMAL:SW_HIDE);

  //int qual=SendDlgItemMessage(hwndDlg,IDC_QUALITY,CB_GETCURSEL,0,0);
  //if (qual == CB_ERR) qual = 0;

  int qual = GetQuality(hwndDlg);

  switch(qual)
  {
  case LQP_R3MIX:
  case LQP_STANDARD:
  case LQP_FAST_STANDARD:
  case LQP_EXTREME:
  case LQP_FAST_EXTREME:
  case LQP_INSANE: 
  case LQP_ABR:
  case LQP_CBR:
    EnableWindow(GetDlgItem(hwndDlg,IDC_STEREOMODE),0);
    EnableWindow(GetDlgItem(hwndDlg,IDC_VBRMETHOD),0);
    break;
  default:
    EnableWindow(GetDlgItem(hwndDlg,IDC_STEREOMODE),1);
    EnableWindow(GetDlgItem(hwndDlg,IDC_VBRMETHOD),1);
    break;
  }
  if (qual == 4) 
  {
    ShowWindow(GetDlgItem(hwndDlg, IDC_VBRQ_TEXT), SW_HIDE);
    ShowWindow(GetDlgItem(hwndDlg, IDC_VBRQ), SW_HIDE);
    //EnableWindow(GetDlgItem(hwndDlg,IDC_VBRQ),0);
  }
}

static void updateVbrOptions(HWND hwndDlg)
{
  //int qual=SendDlgItemMessage(hwndDlg,IDC_QUALITY,CB_GETCURSEL,0,0);
  //if (qual == CB_ERR) qual = 0;
  int qual = GetQuality(hwndDlg);
  switch(qual)
  {
  case LQP_R3MIX:
    SendDlgItemMessage(hwndDlg,IDC_VBRMETHOD,CB_SETCURSEL,3,0);
    setBitrates(hwndDlg,96,224);
    SendDlgItemMessage(hwndDlg,IDC_STEREOMODE,CB_SETCURSEL,1,0);
    SendDlgItemMessage(hwndDlg,IDC_VBRQ,CB_SETCURSEL,1,0);
    break;
  case LQP_STANDARD:
    SendDlgItemMessage(hwndDlg,IDC_VBRMETHOD,CB_SETCURSEL,2,0);
    setBitrates(hwndDlg,128,320);
    SendDlgItemMessage(hwndDlg,IDC_STEREOMODE,CB_SETCURSEL,1,0);
    SendDlgItemMessage(hwndDlg,IDC_VBRQ,CB_SETCURSEL,2,0);
    break;
  case LQP_FAST_STANDARD:
    SendDlgItemMessage(hwndDlg,IDC_VBRMETHOD,CB_SETCURSEL,3,0);
    setBitrates(hwndDlg,128,320);
    SendDlgItemMessage(hwndDlg,IDC_STEREOMODE,CB_SETCURSEL,1,0);
    SendDlgItemMessage(hwndDlg,IDC_VBRQ,CB_SETCURSEL,2,0);
    break;
  case LQP_EXTREME:
    SendDlgItemMessage(hwndDlg,IDC_VBRMETHOD,CB_SETCURSEL,2,0);
    setBitrates(hwndDlg,128,320);
    SendDlgItemMessage(hwndDlg,IDC_STEREOMODE,CB_SETCURSEL,1,0);
    SendDlgItemMessage(hwndDlg,IDC_VBRQ,CB_SETCURSEL,2,0);
    break;
  case LQP_FAST_EXTREME:
    SendDlgItemMessage(hwndDlg,IDC_VBRMETHOD,CB_SETCURSEL,3,0);
    setBitrates(hwndDlg,128,320);
    SendDlgItemMessage(hwndDlg,IDC_STEREOMODE,CB_SETCURSEL,1,0);
    SendDlgItemMessage(hwndDlg,IDC_VBRQ,CB_SETCURSEL,2,0);
    break;
  case LQP_INSANE:
    SendDlgItemMessage(hwndDlg,IDC_VBRMETHOD,CB_SETCURSEL,0,0);
    setBitrates(hwndDlg,320,320);
    SendDlgItemMessage(hwndDlg,IDC_STEREOMODE,CB_SETCURSEL,1,0);
    break;
  case LQP_ABR:
    SendDlgItemMessage(hwndDlg,IDC_VBRMETHOD,CB_SETCURSEL,5,0);
    setBitrates(hwndDlg,64,320);
    SendDlgItemMessage(hwndDlg,IDC_STEREOMODE,CB_SETCURSEL,1,0);
    SendDlgItemMessage(hwndDlg,IDC_VBRQ,CB_SETCURSEL,2,0);
    break;
  case LQP_CBR:
    SendDlgItemMessage(hwndDlg,IDC_VBRMETHOD,CB_SETCURSEL,0,0);
    SendDlgItemMessage(hwndDlg,IDC_STEREOMODE,CB_SETCURSEL,1,0);
    break;
  }

  updateOptions(hwndDlg);
}

static int LoadDefaultConfig(void **data, const char *desc)
{
  static WDL_HeapBuf m_hb;
  const char *fn=get_ini_file();
  int l=GetPrivateProfileInt(desc,"default_size",0,fn);
  if (l<1) return 0;
  
  if (GetPrivateProfileStruct(desc,"default",m_hb.Resize(l),l,fn))
  {
    *data = m_hb.Get();
    return l;
  }
  return 0;
}
static void SaveDefaultConfig(void *data, int l, const char *desc)
{
  const char *fn=get_ini_file();
  char buf[64];
  sprintf(buf,"%d",l);
  WritePrivateProfileString(desc,"default_size",buf,fn);
  WritePrivateProfileStruct(desc,"default",data,l,fn);
}

WDL_DLGRET wavecfgDlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  switch (uMsg)
  {
    case WM_INITDIALOG:
      {
        void *cfgdata=((void **)lParam)[0];
        int cfgdata_l=(int)((void**)lParam)[1];
        int br=128, vbrq=2, abr=128, vbrmax=320, quality=6, stereomode=1, vbrmethod=-1;
        if (cfgdata_l < 32 || *((int *)cfgdata) != REAPER_FOURCC('m','p','3','l'))
          cfgdata_l=LoadDefaultConfig(&cfgdata,"mp3 encoder defaults");

        if (cfgdata_l >= 32 && *((int *)cfgdata) == REAPER_FOURCC('m','p','3','l'))
        {
          br=REAPER_MAKELEINT(((int *)(((unsigned char *)cfgdata)+4))[0]);
          stereomode=REAPER_MAKELEINT(((int *)(((unsigned char *)cfgdata)+4))[1]);
          quality=REAPER_MAKELEINT(((int *)(((unsigned char *)cfgdata)+4))[2]);
          vbrmethod=REAPER_MAKELEINT(((int *)(((unsigned char *)cfgdata)+4))[3]);
          vbrq=REAPER_MAKELEINT(((int *)(((unsigned char *)cfgdata)+4))[4]);
          vbrmax=REAPER_MAKELEINT(((int *)(((unsigned char *)cfgdata)+4))[5]);
          abr=REAPER_MAKELEINT(((int *)(((unsigned char *)cfgdata)+4))[6]);
        }

        int i;
        for(i=0;m_bitrates[i];i++)
        {
          char buf[64];
          wsprintf(buf, "%d", m_bitrates[i]);
          SendDlgItemMessage(hwndDlg,IDC_BITRATE   ,CB_ADDSTRING,0,(LPARAM)buf);
          SendDlgItemMessage(hwndDlg,IDC_MAXBITRATE,CB_ADDSTRING,0,(LPARAM)buf);
          SendDlgItemMessage(hwndDlg,IDC_AVGBITRATE,CB_ADDSTRING,0,(LPARAM)buf);
          SendDlgItemMessage(hwndDlg,IDC_BITRATE   ,CB_SETITEMDATA,i,m_bitrates[i]);
          SendDlgItemMessage(hwndDlg,IDC_MAXBITRATE,CB_SETITEMDATA,i,m_bitrates[i]);
          SendDlgItemMessage(hwndDlg,IDC_AVGBITRATE,CB_SETITEMDATA,i,m_bitrates[i]);
        }

        for(i=0;m_bitrates[i];i++)
        {
          if (m_bitrates[i] == br) SendDlgItemMessage(hwndDlg,IDC_BITRATE,CB_SETCURSEL,i,0);
          if (m_bitrates[i] == abr) SendDlgItemMessage(hwndDlg,IDC_AVGBITRATE,CB_SETCURSEL,i,0);
          if (m_bitrates[i] == vbrmax) SendDlgItemMessage(hwndDlg,IDC_MAXBITRATE,CB_SETCURSEL,i,0);
        }

/*
// need to preserve numbering
0        SendDlgItemMessage(hwndDlg,IDC_QUALITY,CB_ADDSTRING,0,(LPARAM)"Normal");
1        SendDlgItemMessage(hwndDlg,IDC_QUALITY,CB_ADDSTRING,0,(LPARAM)"Low");
2        SendDlgItemMessage(hwndDlg,IDC_QUALITY,CB_ADDSTRING,0,(LPARAM)"High");
3        SendDlgItemMessage(hwndDlg,IDC_QUALITY,CB_ADDSTRING,0,(LPARAM)"Voice");
4        SendDlgItemMessage(hwndDlg,IDC_QUALITY,CB_ADDSTRING,0,(LPARAM)"R3mix");
5        SendDlgItemMessage(hwndDlg,IDC_QUALITY,CB_ADDSTRING,0,(LPARAM)"Very High");
6        SendDlgItemMessage(hwndDlg,IDC_QUALITY,CB_ADDSTRING,0,(LPARAM)"--alt-preset standard");
7        SendDlgItemMessage(hwndDlg,IDC_QUALITY,CB_ADDSTRING,0,(LPARAM)"--alt-preset fast standard");
8        SendDlgItemMessage(hwndDlg,IDC_QUALITY,CB_ADDSTRING,0,(LPARAM)"--alt-preset extreme");
9        SendDlgItemMessage(hwndDlg,IDC_QUALITY,CB_ADDSTRING,0,(LPARAM)"--alt-preset fast extreme");
10       SendDlgItemMessage(hwndDlg,IDC_QUALITY,CB_ADDSTRING,0,(LPARAM)"--alt-preset insane");
11       SendDlgItemMessage(hwndDlg,IDC_QUALITY,CB_ADDSTRING,0,(LPARAM)"--alt-preset abr");
12       SendDlgItemMessage(hwndDlg,IDC_QUALITY,CB_ADDSTRING,0,(LPARAM)"--alt-preset cbr");

         SendDlgItemMessage(hwndDlg,IDC_QUALITY,CB_SETCURSEL, quality,0);
*/

        SetQualityStr(hwndDlg, "Insane (best)", 10);
        SetQualityStr(hwndDlg, "Extreme", 8);
        SetQualityStr(hwndDlg, "Fast extreme", 9);
        SetQualityStr(hwndDlg, "Standard", 6);
        SetQualityStr(hwndDlg, "Fast standard", 7);
        SetQualityStr(hwndDlg, "  ABR (manual)", 11);
        SetQualityStr(hwndDlg, "  CBR (manual)", 12);
        SetQualityStr(hwndDlg, "    Custom (normal)", 0);
        SetQualityStr(hwndDlg, "    Custom (low)", 1);
        SetQualityStr(hwndDlg, "    Custom (high)", 2);
        SetQualityStr(hwndDlg, "    Custom (voice)", 3);
        SetQualityStr(hwndDlg, "    Custom (r3mix)", 4);
        SetQualityStr(hwndDlg, "    Custom (very high)", 5);

        SetQuality(hwndDlg, quality);

        for (i=0;i<10;i++)
        {
          char buf[64];
          wsprintf(buf,"%d%s",i,i==0?"(High)":i==9?"(Low)":"");
          SendDlgItemMessage(hwndDlg,IDC_VBRQ,CB_ADDSTRING,0,(LPARAM)buf);      
        }
        SendDlgItemMessage(hwndDlg,IDC_VBRQ,CB_SETCURSEL,vbrq,0);

        // -1=none, 0=default, 1=old, 2=new, 3=mtrh, 4=abr
        SendDlgItemMessage(hwndDlg,IDC_VBRMETHOD,CB_ADDSTRING,0,(LPARAM)"CBR");
        SendDlgItemMessage(hwndDlg,IDC_VBRMETHOD,CB_ADDSTRING,0,(LPARAM)"VBR default");
        SendDlgItemMessage(hwndDlg,IDC_VBRMETHOD,CB_ADDSTRING,0,(LPARAM)"VBR old");
        SendDlgItemMessage(hwndDlg,IDC_VBRMETHOD,CB_ADDSTRING,0,(LPARAM)"VBR new");
        SendDlgItemMessage(hwndDlg,IDC_VBRMETHOD,CB_ADDSTRING,0,(LPARAM)"VBR mtrh");
        SendDlgItemMessage(hwndDlg,IDC_VBRMETHOD,CB_ADDSTRING,0,(LPARAM)"ABR");
        SendDlgItemMessage(hwndDlg,IDC_VBRMETHOD,CB_SETCURSEL,vbrmethod+1,0);

        //0=stereo,1=jstereo,2=mchannel,3=mono
        SendDlgItemMessage(hwndDlg,IDC_STEREOMODE,CB_ADDSTRING,0,(LPARAM)"Stereo");
        SendDlgItemMessage(hwndDlg,IDC_STEREOMODE,CB_ADDSTRING,0,(LPARAM)"Joint Stereo");
        SendDlgItemMessage(hwndDlg,IDC_STEREOMODE,CB_ADDSTRING,0,(LPARAM)"Multi-Channel");
        SendDlgItemMessage(hwndDlg,IDC_STEREOMODE,CB_ADDSTRING,0,(LPARAM)"Mono");
        SendDlgItemMessage(hwndDlg,IDC_STEREOMODE,CB_SETCURSEL,stereomode,0);

        updateOptions(hwndDlg);
      }
    return 0;
    case WM_COMMAND:
      if (LOWORD(wParam) == IDC_VBRMETHOD && HIWORD(wParam) == CBN_SELCHANGE)
        updateOptions(hwndDlg);
      if (LOWORD(wParam) == IDC_QUALITY && HIWORD(wParam) == CBN_SELCHANGE)
        updateVbrOptions(hwndDlg);
      break;
    case WM_USER+1024:
      {
        if (wParam) *((int *)wParam)=32;
        if (lParam)
        {
          int br = SendDlgItemMessage(hwndDlg, IDC_BITRATE, CB_GETITEMDATA, SendDlgItemMessage(hwndDlg,IDC_BITRATE,CB_GETCURSEL,0,0), 0);
          int stereomode = SendDlgItemMessage(hwndDlg,IDC_STEREOMODE,CB_GETCURSEL,0,0);
          //int quality = SendDlgItemMessage(hwndDlg,IDC_QUALITY,CB_GETCURSEL,0,0);
          int quality = GetQuality(hwndDlg);
          int vbr_method = SendDlgItemMessage(hwndDlg,IDC_VBRMETHOD,CB_GETCURSEL,0,0);
          if (vbr_method == CB_ERR) vbr_method = 0;
          int vbrq = SendDlgItemMessage(hwndDlg,IDC_VBRQ,CB_GETCURSEL,0,0);
          int vbrmax = SendDlgItemMessage(hwndDlg, IDC_MAXBITRATE, CB_GETITEMDATA, SendDlgItemMessage(hwndDlg,IDC_MAXBITRATE,CB_GETCURSEL,0,0), 0);
          int abr = SendDlgItemMessage(hwndDlg, IDC_AVGBITRATE, CB_GETITEMDATA, SendDlgItemMessage(hwndDlg,IDC_AVGBITRATE,CB_GETCURSEL,0,0), 0);

          ((int *)lParam)[0] = REAPER_FOURCC('m','p','3','l');
          ((int *)(((unsigned char *)lParam)+4))[0]=REAPER_MAKELEINT(br);
          ((int *)(((unsigned char *)lParam)+4))[1]=REAPER_MAKELEINT(stereomode);
          ((int *)(((unsigned char *)lParam)+4))[2]=REAPER_MAKELEINT(quality);
          ((int *)(((unsigned char *)lParam)+4))[3]=REAPER_MAKELEINT(vbr_method-1);
          ((int *)(((unsigned char *)lParam)+4))[4]=REAPER_MAKELEINT(vbrq);
          ((int *)(((unsigned char *)lParam)+4))[5]=REAPER_MAKELEINT(vbrmax);
          ((int *)(((unsigned char *)lParam)+4))[6]=REAPER_MAKELEINT(abr);
        }
      }
    return 0;
    case WM_DESTROY:
      {
        char buf[32];
        SendMessage(hwndDlg,WM_USER+1024,0,(LPARAM)buf);
        SaveDefaultConfig(buf,32,"mp3 encoder defaults");
      }
    return 0;
  }
  return 0;
}

static HWND ShowConfig(const void *cfg, int cfg_l, HWND parent)
{
  if (cfg_l >= 4 && *((int *)cfg) == REAPER_FOURCC('m','p','3','l')) 
  {
    const void *x[2]={cfg,(void *)(INT_PTR)cfg_l};
    return CreateDialogParam(g_hInst,MAKEINTRESOURCE(IDD_MP3SINK_CFG),parent,wavecfgDlgProc,(LPARAM)x);
  }
  return 0;
}

#ifdef __APPLE__
#define MP3LIB "libmp3lame.dylib"
#define REAPERAPP "REAPER.app"
#else
#define MP3LIB "lame_enc.dll"
#define REAPERAPP "REAPER.exe"
#endif

WDL_DLGRET MP3DlgProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
  switch (msg)
  {
    case WM_INITDIALOG:
    {
      const char* msg = "Because of patent and licensing issues, REAPER cannot encode MP3 audio natively."
        "\r\n\r\n"
        MP3LIB " is a freely available encoder written by The LAME Project."
        " The LAME MP3 encoder is considered one of the best encoders available."
        " \r\n\r\n"
        "Place " MP3LIB " in the same directory as " REAPERAPP " and restart REAPER in order to enable"
        " recording and rendering directly to MP3.";
      SetDlgItemText(hwndDlg, IDC_LAME_MSG, msg);

      SetDlgItemText(hwndDlg, IDC_LAME_LINK, "Download " MP3LIB);
    }
    return 0;

    case WM_COMMAND:
      switch (LOWORD(wParam))
      {
        case IDC_LAME_LINK:
        {
          const char* url = "http://www.google.com/search?&q=" MP3LIB;
          ShellExecute(hwndDlg, "open", url,"","",0);
        }
        return 0;

        case IDOK:
        {
          EndDialog(hwndDlg, 1);
        }
        return 0;
      }
  }
  return 0;
}

static PCM_sink *CreateSink(const char *filename, void *cfg, int cfg_l, int nch, int srate, bool buildpeaks)
{
  if (cfg_l >= 4 && *((int *)cfg) == REAPER_FOURCC('m','p','3','l')) 
  {
    if (!LameEncoder::CheckDLL())
    {
      DialogBox(g_hInst, MAKEINTRESOURCE(IDD_GET_LAME), g_main_hwnd, MP3DlgProc);
      return 0;
    }
    PCM_sink_mp3lame *v=new PCM_sink_mp3lame(filename,cfg,cfg_l,nch,srate,buildpeaks);
    if (v->IsOpen()) return v;
    delete v;
  }
  return 0;
}


static int ExtendedSinkInfo(int call, void* parm1, void* parm2, void* parm3)
{
  if (call == PCMSINKEXT_GETFORMATDESC||call==PCMSINKEXT_GETFORMATDATARATE)
  {
    void* cfg = parm1;
    int len = (int)parm2;
    if (len<4 || *((int*)cfg) != REAPER_FOURCC('m','p','3','l')) return 0;

    int br = 128;
    int vbrm = -1;
    int abr = 128;

    char* desc = (char*)parm3;
    if (len >= 32) 
    {
      br=REAPER_MAKELEINT(((int *)(((unsigned char *)cfg)+4))[0]);
      vbrm=REAPER_MAKELEINT(((int *)(((unsigned char *)cfg)+4))[3]);
      abr=REAPER_MAKELEINT(((int *)(((unsigned char *)cfg)+4))[6]);     
    }   
    if (call==PCMSINKEXT_GETFORMATDATARATE)
    {
      if (vbrm >= 0) return (abr*1000*3/2)/8;
      
      return (br*1000)/8;
    }

    if (vbrm < 0) sprintf(desc,"%dkbps MP3",br);
    else sprintf(desc,"VBR MP3");
    return 1;

  }
  return 0;
}

pcmsink_register_ext_t mySinkRegStruct={{GetFmt,GetExtension,ShowConfig,CreateSink},ExtendedSinkInfo};